// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- // 
// C++ Source Code File Name: btree2.cpp 
// Compiler Used: MSVC40, DJGPP 2.7.2.1, GCC 2.7.2.1, HP CPP 10.24
// Produced By: Doug Gaer   
// File Creation Date: 02/12/1997 
// Date Last Modified: 04/09/1999
// Copyright (c) 1997 Dougas M. Gaer
// ----------------------------------------------------------- // 
// ------------- Program Description and Details ------------- // 
// ----------------------------------------------------------- // 
/*
This test program is used to store multiple B-trees in a single
index file.

The VBD C++ classes are copyright (c) 1997, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.
*/
// ----------------------------------------------------------- // 
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include "btree.h"
#include "btwalk.h"
#include "btreeprt.h"
#include "vbdstats.h"

void pause()
{
  cout << endl;
  cout << "Press enter to continue..." << endl;
  cin.get();
}

int main()
{
  EntryKey e;
  int cache_size, status1, status2, i, j;
  
  // Cache size for the B-tree
  cache_size = 15;

  // Balanced multi-way file-base tree index file extention
 const char *fname = "testfile.btx";
  
  cout << "Making B-tree index file with cache size of: "
       << cache_size << endl;
  cout << endl;
  
  VBDFilePtr f(new VBDFile);
  Btree btx1(cache_size);
  Btree btx2(cache_size);
  
  // Create the file with room for two B-tree headers
  // in the static area
  f->Create(fname, 2*sizeof(BtreeHeader));

  // Connect the first tree, create the tree, with a header
  // at the location right after the VBD file header.
  btx1.Connect(f, 1, sizeof(FileHeader));
  
  // Connect the second tree, create the tree, with a header
  // at the location right after the first tree's header.
  btx2.Connect(f, 1, sizeof(FileHeader)+sizeof(BtreeHeader));
  
  // Data for the first tree
  char *aa = "DOG";
  char *bb = "CAT";
  char *cc = "FISH";
  char *dd = "MOUSE";
  char *ee = "BIRD";
  char *ff = "PIG";
  char *gg = "HORSE";
  char *hh = "LION";
  char *ii = "SNAKE";
  char *jj = "COW";
  char *kk = "ARMADILLO";
  char *ll = "GROUPER";
  char *mm = "RAT";

  // Data for the second tree
  char *nn = "MONKEY";
  char *oo = "ZEBRA";
  char *pp = "STARFISH";
  char *qq = "LIZARD";
  char *rr = "CRAB";
  char *ss = "SNAIL";
  char *tt = "GORILLA";
  char *uu = "LOBSTER";
  char *vv = "TURKEY";
  char *ww = "BEETLE";
  char *xx = "SHARK";
  char *yy = "CLAM";
  char *zz = "OYSTER";
    
  const int NKEYS = 13;
  
  char *keys1[NKEYS] = { aa, bb, cc, dd, ee, ff, gg, hh, ii, jj,
			 kk, ll, mm };

  char *keys2[NKEYS] = { nn, oo, pp, qq, rr, ss, tt, uu, vv, ww,
			 xx, yy, zz};

  char exists1[NKEYS];
  char exists2[NKEYS];
  
  LineCount = 0;

  for(i = 0, j = 0; i < NKEYS; i++, j++) {
    // Adding keys to the first and second tree
    status1 = btx1.Add(keys1[i]);
    status2 = btx2.Add(keys2[j]);
    LineCount++;
    if (status1 != 1 || status2 != 1) {
      cout << endl << "Problem adding keys" << endl;
      Error->SignalException(EHandler::AssertError);      
    }
    else {
      if(LineCount > DisplayLines) {
	pause();
	LineCount = 0;
      }
      cout << "Adding key " << keys1[i] << " to B-tree #1" << endl;
      cout << "Adding key " << keys2[j] << " to B-tree #2" << endl;
    }
    exists1[i] = 1;
    exists2[j] = 1;
  }

  LineCount = 0;
  cout << endl << "Indexing is complete" << endl;
  pause();
  
  cout << "Displaying stats for B-tree #1" << endl;
  BTreeStats(btx1);
  pause();

  cout << "Displaying stats for B-tree #2" << endl;
  BTreeStats(btx2);
  pause();

  VBDStats(f);
  pause();

  cout << "Walking through the B-tree #1" << endl;
  cout << endl;
  BtreeWalk(btx1.GetRoot(), PrintNode, &btx1);
  pause();

  cout << "Walking through the B-tree #2" << endl;
  cout << endl;
  BtreeWalk(btx2.GetRoot(), PrintNode, &btx2);
  pause();
  
  cout << "Listing B-tree #1" << endl;
  pause();
  LineCount = 0;
  ListInOrder(btx1.GetRoot(), &btx1);

  cout << "Listing B-tree #2" << endl;
  pause();
  LineCount = 0;
  ListInOrder(btx2.GetRoot(), &btx2);

  cout << "Deleting all entries in B-tree #1" << endl;
  pause();
  for(i = 0; i < NKEYS; i++) {
    if(btx1.Remove(keys1[i]) != 1) {
      cout << endl << "Problem removing " << keys1[i] << " - " << i << endl;
      Error->SignalException(EHandler::AssertError);
    }
    exists1[i] = 0;
  }

  cout << "Displaying stats for B-tree #1" << endl;
  BTreeStats(btx1);
  pause();

  cout << "Displaying stats for B-tree #2" << endl;
  BTreeStats(btx2);
  pause();

  cout << "Printing the B-tree #1" << endl;
  LineCount = 0;
  PrintTree(btx1.GetRoot(), &btx1);
  pause();
  
  cout << "Printing the B-tree #2" << endl;
  LineCount = 0;
  PrintTree(btx2.GetRoot(), &btx2);
  pause();
  
  VBDStats(f);
  
  return 0;
}
// ----------------------------------------------------------- //
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //
