// ------------------------------- //
// -------- Start of File -------- //
// ------------------------------- //
// ----------------------------------------------------------- // 
// C++ Header File Name: config.h
// Compiler Used: MSVC40, DJGPP 2.7.2.1, GCC 2.7.2.1, HP CPP 10.24
// Produced By: Doug Gaer  
// File Creation Date: 02/19/1996 
// Date Last Modified: 12/13/1999
// Copyright (c) 1997 Douglas M. Gaer
// ----------------------------------------------------------- // 
// ---------- Include File Description and Details  ---------- // 
// ----------------------------------------------------------- // 
/*
The VBD C++ classes are copyright (c) 1997, by Douglas M. Gaer.
All those who put this code or its derivatives in a commercial
product MUST mention this copyright in their documentation for
users of the products in which this code or its derivative
classes are used. Otherwise, you have the freedom to redistribute
verbatim copies of this source code, adapt it to your specific
needs, or improve the code and release your improvements to the
public provided that the modified files carry prominent notices
stating that you changed the files and the date of any change.

THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
THE ENTIRE RISK OF THE QUALITY AND PERFORMANCE OF THIS SOFTWARE
IS WITH YOU. SHOULD ANY ELEMENT OF THIS SOFTWARE PROVE DEFECTIVE,
YOU WILL ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR, OR
CORRECTION.

The Config class is a character string based singly linked list
that is used to load program parameters from a ASCII text based
configuration file.

Changes:
================================================================
09/09/1999 - Several changes made to corrected memory leak. The
ConfigData class is no longer used. Config data nodes are now
user defined string objects instead of null terminated C
strings. The ConfigListNode and ConfigList classes have been
added. The singly list functions were not incorporated in the
existing classes to maintain compatibility with previous
releases.
Changed by: Doug Gaer

09/09/1999 - Added the Config::GetFloatValue() function to
process floating point values. The GetDFPValue() was retained
to maintain compatibility with previous releases.
Added by: Doug Gaer

09/09/1999 - Added const versions of Load(), GetDFPValue(),
GetStrValue(), GetIntValue(), and GetLongValue().
Added by: Doug Gaer

09/10/1999 - Added functions to write config values to the
configuration file and change configuration values.
Added by: Doug Gaer

09/10/1999 - Added the ReLoad() function to reload the config
file after a write operation.
Added by: Doug Gaer

09/10/1999 - Added UString, char *, and const char * overloads
of all read, write, and open functions.
Added by: Doug Gaer

09/17/1999 - Added the Config::SetParmID() function to enable an
application to change the parameter ID string. The Config::GetParmID()
function will return the current parameter ID string.
Added by: Doug Gaer

09/17/1999 - The Config class inherits the vConfigList class 
as a public base class instead of a private base class. This
change was implemented to allow Config objects to use the
public member functions of the VConfigList class.
Changed by: Doug Gaer

09/17/1999 - Added new Config constructors to allow the application
to set the parameter ID string when the object is created.
Added by: Doug Gaer

09/30/1999 - Added Config::GetCommentChar() and Config::SetCommentChar()
function to allow specific characters to identify comments in the config
file. The configuration manager will use a number sign '#' by default.
Added by: Doug Gaer

09/30/1999 - Added Config::FilterComments() and Config::ReadComments()
function to allow an application to read comment strings or ignore them.
All comments will be ignored by default.
Added by:Doug Gaer

09/30/1999 - Added the Config::WriteCommentLine() and Config::WriteLine()
function to allow the application to generate config files with
comment strings and custom text.
Added by: Doug Gaer

09/30/1999 - Modified all versions of Load(), GetDFPValue(),
GetStrValue(), GetIntValue(), and GetLongValue(). The user
now has the option to search for parameter names anywhere in the
parameter string.
Changed by: Doug Gaer

10/04/1999 - Added unsigned overloads for Config::ChangeConfigValue()
function.
Added by: Doug Gaer

12/13/1999 - Added the Config::FilterLineFeed() function to
remove any line feeds from a line of text read from a file.
Added by: Doug Gaer
*/
// ----------------------------------------------------------- // 
#ifndef __CONFIG_HPP__
#define __CONFIG_HPP__

#include "sllistb.h"
#include "ustring.h"

// Configuration data nodes 
class vConfigListNode: public SNodeBase
{
public:
  vConfigListNode() { } 
  vConfigListNode(const UString &X) : Data(X) { } 

public:
  vConfigListNode *GetNext() { return (vConfigListNode *)Next; }
  const vConfigListNode *GetNext() const { return (vConfigListNode *)Next; }

public:
  UString Data;
};

// Singly linked list used to store configuration data
class vConfigList : public SLListBase
{
public:
  vConfigList() { }
  virtual ~vConfigList();
  vConfigList(const vConfigList &X) { Copy(X); }
  void operator=(const vConfigList &X) { Copy(X); }

public:
  int Copy(const vConfigList &List);
  const vConfigListNode *Find(const UString &X,
			      const vConfigListNode *ptr=0) const;
  vConfigListNode *Find(const UString &X, vConfigListNode *ptr=0);
  vConfigListNode *FindAny(const UString &X, vConfigListNode *ptr=0);
  vConfigListNode *Store(const UString &X);
  vConfigListNode *GetFront();
  const vConfigListNode *GetFront() const;
  int IsHeader(const vConfigListNode *Node) const;
  
protected:
  virtual vConfigListNode *AllocNode(const UString &X);
  virtual SNodeBase *DupNode(const SNodeBase *Node);
  virtual void FreeNode(SNodeBase *Node);
};

// Configuration manager class
class Config : public vConfigList
{
public:
  Config() {
    FileName = "\0"; parm_ID = "="; comment_char = '#';
    filter_comments = 1;
  }
  Config(const UString &fname) {
    FileName = fname; parm_ID = "="; comment_char = '#';
    filter_comments = 1;
  }
  Config(char *fname) {
    FileName = fname; parm_ID = "="; comment_char = '#';
    filter_comments = 1;
  }
  Config(const char *fname) {
    FileName = fname; parm_ID = "="; comment_char = '#';
    filter_comments = 1;
  }

  // Constructors add to set the parameter ID string: 09/17/1999
  Config(const UString &fname, const UString &p_id) {
    FileName = fname; parm_ID = p_id;
  }
  Config(const UString &fname, char *p_id) {
    FileName = fname; parm_ID = p_id;
  }
  Config(const UString &fname, const char *p_id) {
    FileName = fname; parm_ID = p_id;
  }
  ~Config();

private:
  Config(const Config &ob) { }         // Disallow copying
  void operator=(const Config &ob) { } // Disallow assignment

public: // Functions used to load and unload config file
  int Load();
  int Load(char *fname); 
  int Load(const char *fname);
  int Load(const UString &fname);
  int ReLoad();
  int ReLoad(char *fname); 
  int ReLoad(const char *fname);
  int ReLoad(const UString &fname);
  void UnLoad();
  char *GetFileName() { return FileName.c_str(); }
  void SetFileName(const UString &s) { FileName = s; }
  void SetFileName(const char *s) { FileName = s; }
  void SetFileName(char *s) { FileName = s; }
  
public: // Formatting functions
  // Treat all characters after the ID string as a parameter value
  void SetParmID(const UString &s) { parm_ID = s; }
  void SetParmID(char *s) { parm_ID = s; }
  void SetParmID(const char *s) { parm_ID = s; }
  char *GetParmID() { return parm_ID.c_str(); }
  char GetCommentChar() { return comment_char; }
  void SetCommentChar(char c) { comment_char = c; }
  void FilterComments() { filter_comments = 1; }
  void ReadComments() { filter_comments = 0; }
  
public: // Functions used to read config file values
  double Config::GetFloatValue(char *Name, int fn=1); // Floating point values
  double Config::GetFloatValue(const char *Name, int fn=1);    
  double Config::GetFloatValue(const UString &Name, int fn=1); 
  double GetDFPValue(char *Name, int fn=1);           // Floating point values
  double GetDFPValue(const char *Name, int fn=1);              
  double GetDFPValue(const UString &Name, int fn=1);           
  char* GetStrValue(char *Name, int fn=1);          // Character string values
  char* GetStrValue(const char *Name, int fn=1);    
  char* GetStrValue(const UString &Name, int fn=1); 
  int GetIntValue(char *Name, int fn=1);            // Integer values
  int GetIntValue(const char *Name, int fn=1);      
  int GetIntValue(const UString &Name, int fn=1);   
  long GetLongValue(char *Name, int fn=1);          // Long integer values
  long GetLongValue(const char *Name, int fn=1);    
  long GetLongValue(const UString &Name, int fn=1); 

public: // Functions used to write config values
  int WriteConfigLine(const UString &parm, const UString &value);
  int WriteConfigLine(const char *parm, const char *value);
  int WriteConfigLine(char *parm, char *value);
  int WriteCommentLine(const UString &s);
  int WriteCommentLine(const char *s);
  int WriteCommentLine(char *s);
  int WriteLine(const UString &s);
  int WriteLine(const char *s);
  int WriteLine(char *s);
  int ChangeConfigValue(const UString &parm, const UString &value);
  int ChangeConfigValue(const char *parm, const char *value);
  int ChangeConfigValue(char *parm, char *value);
  int ChangeConfigValue(const UString &parm, int value);
  int ChangeConfigValue(const char *parm, int value);
  int ChangeConfigValue(char *parm, int value);
  int ChangeConfigValue(const UString &parm, unsigned value);
  int ChangeConfigValue(const char *parm, unsigned value);
  int ChangeConfigValue(char *parm, unsigned value);
  int ChangeConfigValue(const UString &parm, long value);
  int ChangeConfigValue(const char *parm, long value);
  int ChangeConfigValue(char *parm, long value);
  int ChangeConfigValue(const UString &parm, float value);
  int ChangeConfigValue(const char *parm, float value);
  int ChangeConfigValue(char *parm, float value);
  int ChangeConfigValue(const UString &parm, double value);
  int ChangeConfigValue(const char *parm, double value);
  int ChangeConfigValue(char *parm, double value);
  
private: // Internal proccessing functions
  void FilterLineFeed(char &c);
  int StoreCfgData(const UString &str);
  int ChangeConfigLine(const UString &string_to_replace,
		       const UString &string_to_insert,
		       int check_case, int append_lines);
  
private:
  UString FileName;    // Name of the configuration file
  UString parm_ID;     // String used indentify parameters in the file
  char comment_char;   // Char used to ID comments in the config file
  int filter_comments; // If ture all comments will be ignored
};

#endif // __CONFIG_HPP__
// ----------------------------------------------------------- //
// ------------------------------- //
// --------- End of File --------- //
// ------------------------------- //

